#include "CommandQueue.h"
#include "Debug.h"

static ZNLinkedList_t * g_zwpkg_queue = NULL;
static uint32_t last_controller_package_time = 0;
#ifdef LOGGING_UART
void zuno_dbgdumpZWPacakge(ZUNOCommandPacket_t * cmd);
#endif

void ZWQIncomingStat(ZUNOCommandPacket_t * pkg){
    if(pkg->src_node == 1){
        last_controller_package_time = millis();
    }
}
bool ZWQPushPackage(ZUNOCommandPacket_t * pkg){

    if(((pkg->flags & ZUNO_PACKETFLAGS_TEST) == 0) && // Sometimes we need a test pakage that ignores this restriction
        (zunoNID() == 0)) { // We are out of network - don't send anything
		#ifdef LOGGING_DBG
		LOGGING_UART.print(millis());
		LOGGING_UART.println(" Package was dropped! NodeID==0");
		#endif
		return false;
		
	}
    ZUNOCommandPacket_t * stored_pck;
    stored_pck = (ZUNOCommandPacket_t *)malloc(sizeof(ZUNOCommandPacket_t)); 
    //YGn vérifié qu'on est bien dans le même espace mémoire que dans loop()

    if (stored_pck == NULL){
    #ifdef LOGGING_DBG
		LOGGING_UART.print(millis());
		LOGGING_UART.println(" Enqueue: Memory OVF1");
		#endif
        return false;
    }
    memcpy(stored_pck, pkg, sizeof(ZUNOCommandPacket_t));
    // We have to allocated command buffer dynamically too
    // add ZAF offset
    stored_pck->cmd = (uint8_t*)malloc(sizeof(uint8_t)*pkg->len + MAX_ZWTRANSPORT_ENCAP);
    if(stored_pck->cmd == NULL){
        free(stored_pck);
        #ifdef LOGGING_DBG
		LOGGING_UART.print(millis());
		LOGGING_UART.println(" Enqueue: Memory OVF2");
		#endif
        return false;
    }
    memset(stored_pck->cmd, 0, MAX_ZWTRANSPORT_ENCAP);
    stored_pck->cmd += MAX_ZWTRANSPORT_ENCAP;
    // copy command data from the ZAF offset
    memcpy(stored_pck->cmd , pkg->cmd, pkg->len);
    znllPushBack(&g_zwpkg_queue, (void*)stored_pck);
    #if defined(LOGGING_DBG) && defined(FULL_LOG)
	LOGGING_UART.print("\n >>> (");
	LOGGING_UART.print(millis());
	LOGGING_UART.print(") ENQUEUED PACKAGE: ");
	zuno_dbgdumpZWPacakge(pkg);
	#endif
    return true;
}
ZUNOCommandPacket_t * ZWQFindPackage(uint8_t dst_id, uint8_t flags, uint8_t cc, uint8_t cmd){
    ZNLinkedList_t *e;
    ZUNOCommandPacket_t * p;

    zunoEnterCritical();
    for(e=g_zwpkg_queue;e; e=e->next){
        p = (ZUNOCommandPacket_t *) e->data;
        if((flags != 0xFF) && ((p->flags & flags)  !=  flags))
            continue;
        if((dst_id != 0xFF) && (p->dst_node != dst_id))
            continue;
        if((cc != 0xFF) && (p->cmd[0]!= cc))
            continue;
        if((cmd != 0xFF) && (p->cmd[1]!= cmd))
            continue;
        zunoExitCritical();
        return p;
    }
    zunoExitCritical();
    return NULL;
}
bool zunoExtractGroupNode(uint8_t g, uint8_t i, ZUnoAssocNode_t * node){
	 return zunoSysCall(ZUNO_SYSFUNC_ASSOCIATION_NODE, 3, g, i, node) == 0;
}

#ifdef YGNPROTECTEEPROMACCESS
uint32_t u32EepromBusyTime = 0;
void setEEpromBusy(boolean bBusy)
{
  if(bBusy) u32EepromBusyTime = millis();
  else u32EepromBusyTime = 0;
}
#endif

bool zunoCheckSystemQueueStatus(uint8_t channel){ //returns true if busy
  
  #ifdef YGNPROTECTEEPROMACCESS
  uint32_t interval = millis();
  boolean bEepromBusy = (interval - u32EepromBusyTime) < 1000 ;
  interval -= g_zuno_sys->rstat_pkgs_hp_time;
  #else
  uint32_t interval = millis() - g_zuno_sys->rstat_pkgs_hp_time;
  boolean bEepromBusy = false;
  #endif
  
  #ifndef YGNRESETBUSYQUEUE
  boolean bInterval = (interval < SYSTEM_PKG_DOMINATION_TIME);
  #ifdef YGNINTERVALSTART
  bInterval = bInterval && (interval > 100);
  #endif
  if  (((channel > 0) && bInterval) || (bEepromBusy))
    return true;
  else
    return (g_zuno_sys->rstat_priority_counts[channel] > 0);
  #else  //YGNRESETBUSYQUEUE
  static uint32_t l_count_P = 0;
  static uint32_t l_count_I = 0;
  static bool l_bResetWarning = false;
  uint32_t count;
  static uint32_t u32FirstTime;
  bool  b_busy = true;

  
  boolean bInterval = (interval < SYSTEM_PKG_DOMINATION_TIME);
  #ifdef YGNINTERVALSTART
  bInterval = bInterval && (interval > 100);
  #endif
 
  bool  b_busyI = (((channel > 0) && (bInterval))|| (bEepromBusy));
  bool  b_busyP =  (g_zuno_sys->rstat_priority_counts[channel] > 0);
  count = l_count_I + l_count_P;
  if (!b_busyI && !b_busyP)
  {
    #if defined(LOGGING_DBG)
    #if defined(YGNLOGBUSYQUEUE)
    if ((count >= 10) &&  ((millis()-u32FirstTime) >= SYSTEM_PKG_DOMINATION_TIME*3))
    {
      LOGGING_UART.print("*** QCH Busy EXIT: #");
      LOGGING_UART.print(l_count_P);
      LOGGING_UART.print("/");
      LOGGING_UART.print(l_count_I);
      LOGGING_UART.print(" since");LOGGING_UART.println(millis()-u32FirstTime);
    }
    l_count_P = 0;
    l_count_I = 0;
    #endif
    l_bResetWarning = false;
    #endif
    u32FirstTime = 0;
    b_busy = false;
  } 
  else 
  {
    if (u32FirstTime == 0) u32FirstTime = millis();
    #if defined(LOGGING_DBG) && defined(YGNLOGBUSYQUEUE)
    if (b_busyI) l_count_I++;
    if (b_busyP) l_count_P++;
    #endif
  }
  if (b_busy) 
  {
    if (((millis()-u32FirstTime) >= 13000) && (!l_bResetWarning))
    {
      l_bResetWarning = true;
      #if defined(LOGGING_DBG) //g_channels_data.update_map
      LOGGING_UART.print(" >>>> Going to force busy queue to FALSE in 2 s, #");
      LOGGING_UART.print(l_count_P);
      LOGGING_UART.print("/");
      LOGGING_UART.println(l_count_I);
      #endif
    }
    if (((millis()-u32FirstTime) >= 15000))
    {
      #if defined(LOGGING_DBG) 
      LOGGING_UART.print(" >>>> !!!!!!! Forced Busy to FALSE, was since:");LOGGING_UART.println(millis()-u32FirstTime);
      l_bResetWarning = false;
      #endif
      g_zuno_sys->rstat_priority_counts[channel] = 0;
      g_zuno_sys->rstat_pkgs_hp_time = 0;
      b_busy = false;
    }
    #if defined(LOGGING_DBG) && defined(YGNLOGBUSYQUEUE)
    count = l_count_I + l_count_P;
    if (((count % 200) == 0) && (count > 20))
      {LOGGING_UART.printf("*** QCH Busy #%d %d /%d since %d",l_count_P,l_count_I,g_zuno_sys->rstat_priority_counts[channel],(millis()-u32FirstTime));
       LOGGING_UART.printf(", maps:%x %x %x %x %x\n",g_Usrchannels_data.update_map,g_Usrchannels_data.request_map,
              g_Usrchannels_data.report_map,g_Usrchannels_data.sys_reports,g_Usrchannels_data.sys_requests);
        }
    // if (((count % 1000) < 20) && (count > 20)) 
    #endif
  }
  return b_busy;
#endif //YGNRESETBUSYQUEUE

}

uint8_t _isToSend(ZUNOCommandPacket_t * p){
  uint8_t u8src_zw_channel = p->src_zw_channel ;
  if (u8src_zw_channel & ZWAVE_CHANNEL_MAPPED_BIT) 
  {
    return 1;
  }
  uint8_t u8dst_zw_channel = p->dst_zw_channel ;
  bool b_plain_assoc = (u8dst_zw_channel == PLAIN_ASSOC_MAP); // It's a plain associtaion 
  u8dst_zw_channel &= ~(PLAIN_ASSOC_MAP); // Remove plain assoc value
  u8src_zw_channel &= ~(ZWAVE_CHANNEL_MAPPED_BIT);
  if(b_plain_assoc &&  ((u8dst_zw_channel != 0) || (u8src_zw_channel  != 0)))
  {
    return 0;
  }
  return 2; 
} //_isToSend
      

void _ZWQSend(ZUNOCommandPacket_t * p){
  #if (defined(LOGGING_DBG) && (defined(YGNLOGOUTGOINGS) || defined(FULL_LOG)))
  uint8_t u8SendBy = 0;
  #endif
  bool b_plain_assoc = (p->dst_zw_channel == PLAIN_ASSOC_MAP); // It's a plain associtaion 
  p->dst_zw_channel &= ~(PLAIN_ASSOC_MAP); // Remove plain assoc value
  if(p->src_zw_channel & ZWAVE_CHANNEL_MAPPED_BIT)
  {
    uint8_t mapped_channel = p->src_zw_channel & ~(ZWAVE_CHANNEL_MAPPED_BIT);
    p->src_zw_channel = 0;
    zunoSysCall(ZUNO_SYSFUNC_SENDPACKET, 1, p);
    p->src_zw_channel = mapped_channel;
    
    #if (defined(LOGGING_DBG) && (defined(YGNLOGOUTGOINGS) || defined(FULL_LOG)))
    u8SendBy = 1;
    LOGGING_UART.print(" >>> (");
    LOGGING_UART.print(millis());
    LOGGING_UART.print(") OUTGOING:-1- ");
    zuno_dbgdumpZWPacakge(p);
    #endif
  }
  if(b_plain_assoc &&  ((p->dst_zw_channel != 0) || (p->src_zw_channel  != 0)))
  {
      return; // do not send association with multichannel encap to plain group
  }
  zunoSysCall(ZUNO_SYSFUNC_SENDPACKET, 1, p); 
  #if (defined(LOGGING_DBG) && (defined(YGNLOGOUTGOINGS) || defined(FULL_LOG)))
  u8SendBy |= 2;
  if (u8SendBy == 2)
  {
    LOGGING_UART.print(" >>> (");
    LOGGING_UART.print(millis());
    LOGGING_UART.print(") OUTGOING:-2- ");
    zuno_dbgdumpZWPacakge(p);
  }
  else if (u8SendBy !=0) LOGGING_UART.println(" >>>>> :!!!!!! OUTGOING twice");
  #endif
}//_ZWQSend

void _ZWQRemovePkg(ZUNOCommandPacket_t * p){
    // revert back cmd addr
    p->cmd -= MAX_ZWTRANSPORT_ENCAP;
    free(p->cmd);
    free(p);
}
bool ZWQIsEmpty(){
    return (znllCount(g_zwpkg_queue) == 0);
}

#if ZUNO_PACKETFLAGS_PRIORITY_MASK > 0xFF
#error "ZUNO_PACKETFLAGS_PRIORITY_MASK size!!!"
#endif

void ZWQProcess(){
    ZNLinkedList_t *e;
    ZUnoAssocNode_t node;
    static uint32_t count_n = 0;
    uint8_t queue_bit_mask;
    
    // Walk through the queue
    int qi;
    int processed_indexes[MAX_PROCESSED_QUEUE_PKGS];
    int processed_indexes_cnt = 0;
    ZUNOCommandPacket_t * p;
    // Process the packages
	int queue_sz = znllCount(g_zwpkg_queue);
    #if defined(LOGGING_DBG) 
      #if defined(FULL_LOG)
      #define LOG_THRLD 0
      #else 
      #define LOG_THRLD 1
      #endif
    if((queue_sz > LOG_THRLD) && ((count_n & 0x3F) == 0)){
	    LOGGING_UART.print(millis());LOGGING_UART.print(" CommandQueue Size:");
        LOGGING_UART.println(queue_sz);
    }
      #undef LOG_THRLD
    #endif
    queue_bit_mask = 0x0;
    for(e=g_zwpkg_queue, qi=0; e!= NULL && qi < queue_sz; e=e->next, qi++){
        p = (ZUNOCommandPacket_t *) e->data;
        #ifdef SYQUEUE_CNT_BARRIER
        uint32_t system_queue_count = g_zuno_sys->rstat_pkgs_queued - g_zuno_sys->rstat_pkgs_processed; //s.pkgs_queued - s.pkgs_processed;
        if(system_queue_count >= MAX_SYS_QUEUE_PKGS){
            #ifdef LOGGING_DBG
		    LOGGING_UART.print("*** SYSTEM QUEUE IS FULL. CNT:");
            LOGGING_UART.println(system_queue_count);
            LOGGING_UART.print("  QUEUED:");
            LOGGING_UART.print(g_zuno_sys->rstat_pkgs_queued);
            LOGGING_UART.print("  PROCESSED:");
            LOGGING_UART.println(g_zuno_sys->rstat_pkgs_processed);
            #endif
            break;
        }
        #endif // SYQUEUE_CNT_BARRIER
        uint8_t q_ch = p->flags & ZUNO_PACKETFLAGS_PRIORITY_MASK; 
        // #ifdef LOGGING_DBG
		    // LOGGING_UART.print("*** QCH:");
        // LOGGING_UART.println(q_ch);
        // #endif

        if ((queue_bit_mask & (0x1 << (q_ch))) != 0x0) {
        //    #ifdef LOGGING_DBG
        //    LOGGING_UART.print("Skip queue: ");
        //    LOGGING_UART.println(q_ch);
        //    #endif
           continue ;
        }
        if(((millis() - last_controller_package_time) < CONTROLLER_INTERVIEW_REQUEST_INTERVAL) && 
           ( q_ch !=  0) &&
           (zunoSecurityStatus() == SECURITY_KEY_S0)){
               #ifdef LOGGING_DBG
                LOGGING_UART.print("Waiting for controller.  QCH:");
                LOGGING_UART.println(q_ch);
                #endif
                queue_bit_mask = queue_bit_mask | (0x1 << (q_ch));
                continue; // Skip report packets during interview
           }

        // Check if we have a free channel to send the package
        if(zunoCheckSystemQueueStatus(q_ch)) {
          //if ((count_n & 0x3F) == 0) {LOGGING_UART.println("Busy Out");count_n++;}
            queue_bit_mask = queue_bit_mask | (0x1 << (q_ch));
            continue; // Needed queue is full for this priority - just skip the package
        }
        
        if(p->flags & ZUNO_PACKETFLAGS_GROUP){
            #if defined(LOGGING_DBG) && defined(FULL_LOG)
		    LOGGING_UART.print(millis());
		    LOGGING_UART.print("*** GROUP:");
            LOGGING_UART.print(p->dst_node);
            LOGGING_UART.print(" I:");
            LOGGING_UART.print(p->dst_zw_channel);
            LOGGING_UART.println(" PKG");
		    #endif
            // Send that packet to group
            // Extract node from association storage
            zunoExtractGroupNode(p->dst_node, p->dst_zw_channel, &node);
            #if defined(LOGGING_DBG) && defined(FULL_LOG)
		    LOGGING_UART.print("*** NODE ID:");
            LOGGING_UART.print(node.dest_nodeid);
            LOGGING_UART.print(" channel:");
            LOGGING_UART.println(node.dest_nodeid);
		    #endif
            if(node.dest_nodeid == 0){
                // There is no more data for this group
                processed_indexes[processed_indexes_cnt++] = qi;
                if(processed_indexes_cnt >= MAX_PROCESSED_QUEUE_PKGS)
                    break;
                continue;
            } 
            // Save group attributes
            uint16_t tmp = p->dst_node;
            uint8_t tmp_ch = p->dst_zw_channel;
            uint8_t tmp_flags = p->flags;
            // Copy Node id of specific device
            p->dst_node = node.dest_nodeid;
            p->dst_zw_channel = node.dest_channel;
            p->zw_rx_secure_opts = node.security_level;
            p->flags &= ~(ZUNO_PACKETFLAGS_GROUP);
            // Send package to target device
            #ifdef YGNSENDONLYONEPKT
            if(_isToSend(p) == 0)
            {
              p->dst_zw_channel &= ~(PLAIN_ASSOC_MAP);
              _ZWQSend(p);
              p->dst_node = tmp;
              p->dst_zw_channel = tmp_ch + 1;
              p->flags = tmp_flags;             
              processed_indexes[processed_indexes_cnt++] = qi;
              if(processed_indexes_cnt >= MAX_PROCESSED_QUEUE_PKGS)
                break;
            }
            else
            #endif
            {
              _ZWQSend(p);
              // Restore group attributes 
              p->dst_node = tmp;
              p->dst_zw_channel = tmp_ch + 1;
              p->flags = tmp_flags;
            }
        } else {
          #if defined(LOGGING_DBG) && defined(FULL_LOG)
            LOGGING_UART.println("Direct _ZWQSend");
          #endif
            _ZWQSend(p);
            processed_indexes[processed_indexes_cnt++] = qi;
            if(processed_indexes_cnt >= MAX_PROCESSED_QUEUE_PKGS)
                break;
        }
    } //for(....
    #if defined(LOGGING_DBG) && defined(FULL_LOG)
    if(processed_indexes_cnt){
	    LOGGING_UART.print("*** PROCESSED:");
        LOGGING_UART.println(processed_indexes_cnt);
    }
	#endif
    // Clean queue
    while(processed_indexes_cnt){
        // The backward order, otherwise it will break the indexes
        qi = processed_indexes[processed_indexes_cnt-1];
        #if defined(LOGGING_DBG) && defined(FULL_LOG)
	    LOGGING_UART.print("*** CLEANUP:");
        LOGGING_UART.println(qi);
	    #endif
        p = (ZUNOCommandPacket_t *)znllRemove(&g_zwpkg_queue, qi);
        _ZWQRemovePkg(p);
        processed_indexes_cnt--;
    }
    count_n++;
}